#include "stdafx.h"
#include "TestWinHTTPWrappers.h"
#include "TestWinHTTPWrappersDlg.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


#ifndef WM_WINHTTP_DOWNLOAD_COMPLETE
#define WM_WINHTTP_DOWNLOAD_COMPLETE WM_APP + 1
#endif //#ifndef WM_WINHTTP_DOWNLOAD_COMPLETE


HRESULT CMyDownloadFileWinHTTPRequest::OnCallback(_In_ HINTERNET hInternet, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength)
{
  CString sMessage;
  switch (dwInternetStatus)
  {
    case WINHTTP_CALLBACK_STATUS_CLOSING_CONNECTION:
    {
      sMessage.Format(_T("Closing the connection to the server, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_CONNECTED_TO_SERVER:
    {
      sMessage.Format(_T("Successfully connected to the server:%ls, Handle:%p\n"), static_cast<LPWSTR>(lpvStatusInformation), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_CONNECTING_TO_SERVER:
    {
      sMessage.Format(_T("Connecting to the server:%ls, Handle:%p\n"), static_cast<LPWSTR>(lpvStatusInformation), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_CONNECTION_CLOSED:
    {
      sMessage.Format(_T("Successfully closed the connection to the server, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_DATA_AVAILABLE:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Data is available to be retrieved, Handle:%p, Data Available:%u\n"), hInternet, *(static_cast<DWORD*>(lpvStatusInformation)));
      break;
    }
    case WINHTTP_CALLBACK_STATUS_HANDLE_CREATED:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Handle created, Handle:%p\n"), *(static_cast<HINTERNET*>(lpvStatusInformation)));
      break;
    }
    case WINHTTP_CALLBACK_STATUS_HANDLE_CLOSING:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Handle closing, Handle:%p\n"), *(static_cast<HINTERNET*>(lpvStatusInformation)));
      break;
    }
    case WINHTTP_CALLBACK_STATUS_HEADERS_AVAILABLE:
    {
      sMessage.Format(_T("The response header has been received, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_INTERMEDIATE_RESPONSE:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Received an intermediate (100 level) status code message from the server, Handle:%p, Status:%u\n"), hInternet, *(static_cast<DWORD*>(lpvStatusInformation)));
      break;
    }
    case WINHTTP_CALLBACK_STATUS_NAME_RESOLVED:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Successfully found the IP address of the server:%ls, Handle:%p\n"), static_cast<LPWSTR>(lpvStatusInformation), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_READ_COMPLETE:
    {
      sMessage.Format(_T("Data was successfully read from the server, Data Read:%u, Handle:%p\n"), dwStatusInformationLength, hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_RECEIVING_RESPONSE:
    {
      sMessage.Format(_T("Waiting for the server to respond to a request, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_REDIRECT:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("An HTTP request is about to automatically redirect the request to %ls, Handle:%p\n"), static_cast<LPWSTR>(lpvStatusInformation), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_REQUEST_ERROR:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      const auto pResult{static_cast<const WINHTTP_ASYNC_RESULT*>(lpvStatusInformation)};
      sMessage.Format(_T("An error occurred while sending an HTTP request, Error:%u, Handle:%p\n"), pResult->dwError, hInternet); //NOLINT(clang-analyzer-core.NullDereference)
      break;
    }
    case WINHTTP_CALLBACK_STATUS_REQUEST_SENT:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Successfully sent the information request to the server, Data Sent:%u, Handle:%p\n"), *(static_cast<DWORD*>(lpvStatusInformation)), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_RESOLVING_NAME:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Looking up the IP address of a server name:%ls, Handle:%p\n"), static_cast<LPWSTR>(lpvStatusInformation), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_RESPONSE_RECEIVED:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Successfully received a response from the server, Data Received:%u, Handle:%p\n"), *(static_cast<DWORD*>(lpvStatusInformation)), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_SECURE_FAILURE:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      const DWORD dwStatusInformation{*(static_cast<DWORD*>(lpvStatusInformation))};
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_CERT_REV_FAILED)
        sMessage.Format(_T("Certification revocation checking has been enabled, but the revocation check failed to verify whether a certificate has been revoked, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_INVALID_CERT)
        sMessage.Format(_T("SSL certificate is invalid, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_CERT_REVOKED)
        sMessage.Format(_T("SSL certificate was revoked, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_INVALID_CA)
        sMessage.Format(_T("The function is unfamiliar with the Certificate Authority that generated the server's certificate, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_CERT_CN_INVALID)
        sMessage.Format(_T("SSL certificate common name (host name field) is incorrect, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_CERT_DATE_INVALID)
        sMessage.Format(_T("SSL certificate date that was received from the server is bad. The certificate is expired, Handle:%p\n"), hInternet);
      if (dwStatusInformation & WINHTTP_CALLBACK_STATUS_FLAG_SECURITY_CHANNEL_ERROR)
        sMessage.Format(_T("The application experienced an internal error loading the SSL libraries, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_SENDING_REQUEST:
    {
      sMessage.Format(_T("Sending the information request to the server, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_SENDREQUEST_COMPLETE:
    {
      sMessage.Format(_T("The request completed successfully, Handle:%p\n"), hInternet);
      break;
    }
    case WINHTTP_CALLBACK_STATUS_WRITE_COMPLETE:
    {
  #pragma warning(suppress: 26496)
      AFXASSUME(lpvStatusInformation != nullptr);
      sMessage.Format(_T("Data was successfully written to the server, Data Written:%u, Handle:%p\n"), *(static_cast<DWORD*>(lpvStatusInformation)), hInternet);
      break;
    }
    default:
    {
      sMessage.Format(_T("Unknown status:%08X, Handle:%p\n"), dwInternetStatus, hInternet);
      break;
    }
  }

  //Update the status static text
#pragma warning(suppress: 26496)
  AFXASSUME(theApp.m_pMainWnd != nullptr);
#pragma warning(suppress: 26489)
  auto pDialog{dynamic_cast<CTestWinHTTPWrappersDlg*>(theApp.m_pMainWnd)};
#pragma warning(suppress: 26496)
  AFXASSUME(pDialog != nullptr);
  ASSERT(pDialog->IsKindOf(RUNTIME_CLASS(CTestWinHTTPWrappersDlg))); //NOLINT(clang-analyzer-core.CallAndMessage)
  pDialog->UpdateStatusText(sMessage);

  //Let the base class do the real work
  return __super::OnCallback(hInternet, dwInternetStatus, lpvStatusInformation, dwStatusInformationLength);
}

HRESULT CMyDownloadFileWinHTTPRequest::OnCallbackComplete(_In_ HRESULT hr, _In_ HINTERNET hInternet, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength)
{
  UNREFERENCED_PARAMETER(hInternet);
  UNREFERENCED_PARAMETER(dwInternetStatus);
  UNREFERENCED_PARAMETER(lpvStatusInformation);
  UNREFERENCED_PARAMETER(dwStatusInformationLength);

  //Release the resources we have been using during the download
  ReleaseResources();

  //Delete a partially downloaded file
  if (hr != S_OK)
    DeleteDownloadedFile();

  //Inform the UI thread that we have finished via a windows message
#pragma warning(suppress: 26496)
  AFXASSUME(theApp.m_pMainWnd != nullptr);
#pragma warning(suppress: 26472)
  theApp.m_pMainWnd->PostMessage(WM_WINHTTP_DOWNLOAD_COMPLETE, 0, static_cast<LPARAM>(hr));

  return S_OK;
}



#pragma warning(suppress: 26455)
CTestWinHTTPWrappersDlg::CTestWinHTTPWrappersDlg(CWnd* pParent) : CDialog{CTestWinHTTPWrappersDlg::IDD, pParent},
                                                                  m_hIcon{AfxGetApp()->LoadIcon(IDR_MAINFRAME)},
                                                                  m_bDownloading{false},
                                                                  m_dwOffset{0},
                                                                  m_nConnection{ConnectionType::UsePreConfig},
                                                                  m_dbLimit{0.0},
                                                                  m_bNoURLRedirect{FALSE},
                                                                  m_sAcceptTypes{_T("*/*")},
                                                                  m_bHTTPPreauthentication{TRUE},
                                                                  m_bProxyPreauthentication{TRUE},
                                                                  m_nHTTPPreauth{0},
                                                                  m_nProxyPreauth{0},
                                                                  m_bAsync{TRUE}
{
}

void CTestWinHTTPWrappersDlg::DoDataExchange(CDataExchange* pDX)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pDX != nullptr);

  //Let the base class do its thing
  __super::DoDataExchange(pDX);

  DDX_Text(pDX, IDC_FILE_DOWNLOAD, m_sFileToDownload);
  DDX_Text(pDX, IDC_FILE_UPLOAD, m_sFileToUpload);
  DDX_Text(pDX, IDC_HTTP_PASSWORD, m_sHTTPPassword);
  DDX_Text(pDX, IDC_PROXY_PASSWORD, m_sProxyPassword);
  DDX_Text(pDX, IDC_URL, m_sURL);
  DDX_Text(pDX, IDC_HTTP_USER, m_sHTTPUserName);
  DDX_Text(pDX, IDC_PROXY_USER, m_sProxyUserName);
  DDX_Text(pDX, IDC_OFFSET, m_dwOffset);
  DDX_CBIndex(pDX, IDC_CONNECTION, m_nConnection);
  DDX_Text(pDX, IDC_PROXYSERVER, m_sProxyServer);
  DDX_Text(pDX, IDC_LIMIT, m_dbLimit);
  DDX_Text(pDX, IDC_VERB, m_sVerb);
  DDV_MinMaxDouble(pDX, m_dbLimit, 0.0, 200000000.0);
  DDX_Check(pDX, IDC_NO_URL_REDIRECT, m_bNoURLRedirect);
  DDX_Text(pDX, IDC_ACCEPT_TYPES, m_sAcceptTypes);
  DDX_Check(pDX, IDC_HTTP_PREAUTHENTICATION, m_bHTTPPreauthentication);
  DDX_Check(pDX, IDC_PROXY_PREAUTHENTICATION, m_bProxyPreauthentication);
  DDX_Text(pDX, IDC_REFERRER, m_sReferrer);
  DDX_CBIndex(pDX, IDC_HTTP_PREAUTH, m_nHTTPPreauth);
  DDX_CBIndex(pDX, IDC_PROXY_PREAUTH, m_nProxyPreauth);
  DDX_Check(pDX, IDC_ASYNC, m_bAsync);
  DDX_Control(pDX, IDC_STATUS, m_wndStatus);

  if (pDX->m_bSaveAndValidate)
  {
    if (m_sURL.IsEmpty())
    {
      AfxMessageBox(_T("Please enter a non-empty URL to download"));
      pDX->PrepareEditCtrl(IDC_URL);
      pDX->Fail();
    }
    if (m_nConnection == UseProxy && m_sProxyServer.GetLength() == 0)
    {
      AfxMessageBox(_T("Please enter a non-empty Proxy Server"));
      pDX->PrepareEditCtrl(IDC_PROXYSERVER);
      pDX->Fail();
    }
  }
}

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CTestWinHTTPWrappersDlg, CDialog) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_PAINT()
  ON_WM_QUERYDRAGICON()
  ON_BN_CLICKED(IDC_DOWNLOAD, &CTestWinHTTPWrappersDlg::OnBnClickedDownload)
  ON_MESSAGE(WM_WINHTTP_DOWNLOAD_COMPLETE, &CTestWinHTTPWrappersDlg::OnDownloadComplete)
END_MESSAGE_MAP()

BOOL CTestWinHTTPWrappersDlg::OnInitDialog()
{
  //Let the base class do its thing
  __super::OnInitDialog();

  //Initialize the status control
  m_wndStatus.LimitText(0);
  SetStatusText(_T("Status of download will appear here\r\n"));

  //Set the icon for this dialog.  The framework does this automatically
  //when the application's main window is not a dialog
  SetIcon(m_hIcon, TRUE); //Set big icon
  SetIcon(m_hIcon, FALSE); //Set small icon

  return TRUE;
}

//If you add a minimize button to your dialog, you will need the code below
//to draw the icon.  For MFC applications using the document/view model,
//this is automatically done for you by the framework.
#pragma warning(suppress: 26434)
void CTestWinHTTPWrappersDlg::OnPaint()
{
  if (IsIconic())
  {
    CPaintDC dc(this); // device context for painting

#pragma warning(suppress: 26490)
    SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

    //Center icon in client rectangle
    const int cxIcon{GetSystemMetrics(SM_CXICON)};
    const int cyIcon{GetSystemMetrics(SM_CYICON)};
    CRect rect;
    GetClientRect(&rect);
    const int x{(rect.Width() - cxIcon + 1) / 2};
    const int y{(rect.Height() - cyIcon + 1) / 2};

    //Draw the icon
    dc.DrawIcon(x, y, m_hIcon);
  }
  else
    __super::OnPaint();
}

//The system calls this function to obtain the cursor to display while the user drags
//the minimized window.
#pragma warning(suppress: 26434 26440)
HCURSOR CTestWinHTTPWrappersDlg::OnQueryDragIcon()
{
#pragma warning(suppress: 26473)
  return static_cast<HCURSOR>(m_hIcon);
}

void CTestWinHTTPWrappersDlg::OnBnClickedDownload()
{
  if (UpdateData(TRUE))
  {
    if (!m_bDownloading)
    {
      SetStatusText(_T("Starting download\r\n"));

      //Crack the URL provided into its constituent parts
      URL_COMPONENTS urlComponents{};
      urlComponents.dwStructSize = sizeof(urlComponents);
#pragma warning(suppress: 26472)
      urlComponents.dwSchemeLength = static_cast<DWORD>(-1);
#pragma warning(suppress: 26472)
      urlComponents.dwHostNameLength = static_cast<DWORD>(-1);
#pragma warning(suppress: 26472)
      urlComponents.dwUrlPathLength = static_cast<DWORD>(-1);
      CStringW sUnicodeURL{m_sURL};
      const BOOL bSuccess{WinHttpCrackUrl(sUnicodeURL.GetBuffer(), 0, 0, &urlComponents)};
      const DWORD dwError{GetLastError()};
      sUnicodeURL.ReleaseBuffer();
      if (!bSuccess)
      {
        CString sMessage;
        sMessage.Format(_T("Failed to parse URL, Error:%u"), dwError);
        UpdateStatusText(sMessage);
        return;
      }

      //Work out the access type we will pass to the "Initialize" method
      DWORD dwAccessType{WINHTTP_ACCESS_TYPE_DEFAULT_PROXY};
      switch (m_nConnection)
      {
        case DirectToInternet:
        {
          dwAccessType = WINHTTP_ACCESS_TYPE_NO_PROXY;
          break;
        }
        case UsePreConfig:
        {
          dwAccessType = WINHTTP_ACCESS_TYPE_DEFAULT_PROXY;
          break;
        }
        case UseProxy:
        {
          dwAccessType = WINHTTP_ACCESS_TYPE_NAMED_PROXY;
          break;
        }
        default:
        {
          ASSERT(FALSE);
          break;
        }
      }

      //Create the session object
      m_session.Close();
#pragma warning(suppress: 26477)
      HRESULT hr = m_session.Initialize(CStringW(AfxGetAppName()), dwAccessType, m_sProxyServer.GetLength() ? CStringW(m_sProxyServer) : WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, m_bAsync ? WINHTTP_FLAG_ASYNC : 0);
      if (FAILED(hr))
      {
        CString sMessage;
        sMessage.Format(_T("Failed to create session, Error:%08X"), hr);
        UpdateStatusText(sMessage);
        return;
      }

      //Create the connection object
      m_connection.Close();
      hr = m_connection.Initialize(m_session, CStringW(urlComponents.lpszHostName, urlComponents.dwHostNameLength), urlComponents.nPort);
      if (FAILED(hr))
      {
        CString sMessage;
        sMessage.Format(_T("Failed to create connection, Error:%08X"), hr);
        UpdateStatusText(sMessage);
        return;
      }

      //Free up out HINTERNET handle if provided
      m_request.Close();

      //Create the request object
      m_request.m_sHTTPUserName = CStringW{m_sHTTPUserName};
      m_request.m_sHTTPPassword = CStringW{m_sHTTPPassword};
      m_request.m_sProxyUserName = CStringW{m_sProxyUserName};
      m_request.m_sProxyPassword = CStringW{m_sProxyPassword};
      m_request.m_nDownloadStartPos = m_dwOffset;
      m_request.m_bHTTPPreauthentication = m_bHTTPPreauthentication;
      m_request.m_dwHTTPPreauthenticationScheme = 1 << m_nHTTPPreauth; //Convert the combo indexes into WINHTTP_AUTH_SCHEME* values
      m_request.m_bProxyPreauthentication = m_bProxyPreauthentication;
      m_request.m_dwProxyPreauthenticationScheme = 1 << m_nProxyPreauth; //Convert the combo indexes into WINHTTP_AUTH_SCHEME* values
      m_request.m_bNoURLRedirect = m_bNoURLRedirect;
      m_request.m_sFileToDownloadInto = CStringW{m_sFileToDownload};
      m_request.m_sFileToUpload = CStringW{m_sFileToUpload};
      m_request.m_dbLimit = m_dbLimit;

      //Tokenize "m_sAcceptTypes" into an array
      std::vector<CStringW> AcceptTypes;
      int nCurPos{0};
      CString sToken{m_sAcceptTypes.Tokenize(_T(",;"), nCurPos)};
      while (sToken != _T(""))
      {
          CStringW sWideToken{sToken};
          AcceptTypes.push_back(sWideToken);
          sToken = m_sAcceptTypes.Tokenize(_T(",;"), nCurPos);
      };
      //Ok, now that we have the string data in an array, convert to the LPCWSR array format
      const size_t nAcceptTypes{AcceptTypes.size()};
      std::vector<LPCWSTR> ppszAcceptTypes{nAcceptTypes + 1, std::allocator<LPCWSTR>{}};
      for (size_t i=0; i<nAcceptTypes; i++)
#pragma warning(suppress: 26446 26481 26489)
        ppszAcceptTypes[i] = AcceptTypes[i].GetString();
#pragma warning(suppress: 26446 26481 26489)
      ppszAcceptTypes[nAcceptTypes] = nullptr;

      //Create the request
#pragma warning(suppress: 26472 26486 26489 26477)
      hr = m_request.Initialize(m_connection, CStringW{urlComponents.lpszUrlPath, static_cast<int>(urlComponents.dwUrlPathLength)},
                                m_sVerb.GetLength() ? CStringW{m_sVerb} : nullptr, nullptr,
                                m_sReferrer.GetLength() ? CStringW{m_sReferrer} : WINHTTP_NO_REFERER,
                                nAcceptTypes ? ppszAcceptTypes.data() : WINHTTP_DEFAULT_ACCEPT_TYPES,
                                urlComponents.nScheme == INTERNET_SCHEME_HTTPS ? WINHTTP_FLAG_SECURE : 0);
      if (FAILED(hr))
      {
        CString sMessage;
        sMessage.Format(_T("Failed to create request, Error:%08X"), hr);
        UpdateStatusText(sMessage);
        return;
      }

      if (m_bAsync)
      {
        //Start the async download
        hr = m_request.SendRequest();
        if (FAILED(hr))
        {
          CString sMessage;
          sMessage.Format(_T("Failed to start download, Error:%08X"), hr);
          UpdateStatusText(sMessage);
          return;
        }

        //Change the button text to "Cancel"
        GetDlgItem(IDC_DOWNLOAD)->SetWindowText(_T("Cancel"));

        //Signal that we are now downloading
        m_bDownloading = true;
      }
      else
      {
        //Do the sync download
        CWaitCursor wait;
        hr = m_request.SendRequestSync();
        if (FAILED(hr))
        {
          CString sMessage;
          bool bValidStatusCode{false};
          const DWORD dwStatusCode{m_request.GetLastStatusCode(bValidStatusCode)};
          if (bValidStatusCode)
            sMessage.Format(_T("Failed to obtain HTTP response, Last HTTP Status Code:%u, Error:%08X"), dwStatusCode, hr);
          else
            sMessage.Format(_T("Failed to obtain HTTP response, Error:%08X"), hr);
          UpdateStatusText(sMessage);
        }
        else
        {
          CString sMessage;
          sMessage.Format(_T("HTTP response retrieved, Time Taken:%I64u ms"), m_request.TimeSinceStartDownload());
          UpdateStatusText(sMessage);
        }
      }
    }
    else
    {
      //Update the status text
      UpdateStatusText(_T("Cancelling download"));

      //Cancel the download
      m_request.Close();
    }
  }
}

void CTestWinHTTPWrappersDlg::SetStatusText(LPCTSTR pszStatus)
{
  m_wndStatus.SetWindowText(pszStatus);
}

void CTestWinHTTPWrappersDlg::UpdateStatusText(LPCTSTR pszStatus)
{
  const int nLength{m_wndStatus.GetWindowTextLength()};
  m_wndStatus.SetSel(nLength, nLength);
  CString sStatus;
  sStatus.Format(_T("%s\r\n"), pszStatus);
  m_wndStatus.ReplaceSel(sStatus);
}

LRESULT CTestWinHTTPWrappersDlg::OnDownloadComplete(WPARAM /*wParam*/, LPARAM lParam)
{
  //Report the error if there was on
#pragma warning(suppress: 26472)
  const auto hr{static_cast<HRESULT>(lParam)};
  if (hr != S_OK)
  {
    CString sMessage;
    bool bValidStatusCode{false};
    const DWORD dwStatusCode{m_request.GetLastStatusCode(bValidStatusCode)};
    if (bValidStatusCode)
      sMessage.Format(_T("Failed to obtain HTTP response, Last HTTP Status Code:%u, Error:%08X"), dwStatusCode, hr);
    else
      sMessage.Format(_T("Failed to obtain HTTP response, Error:%08X"), hr);
    UpdateStatusText(sMessage);
  }
  else
  {
    CString sMessage;
    sMessage.Format(_T("HTTP response retrieved, Time Taken:%I64u ms"), m_request.TimeSinceStartDownload());
    UpdateStatusText(sMessage);
  }

  //Reset the text of the "Download" button
  GetDlgItem(IDC_DOWNLOAD)->SetWindowText(_T("Download"));

  m_bDownloading = false;

  return 0L;
}
